#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: main.modules.cli.application
      @file: argument_parser.py
   @created: Thu, 14 Feb 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import sys
from argparse import ArgumentError, ArgumentParser

class HSArgumentParser(ArgumentParser):

    def _check_value(self, action, value):
        if action.choices is not None and value not in action.choices:
            msg = f"invalid choice: {value} (choose from [{', '.join(map(repr, action.choices))}])"
            raise ArgumentError(action, msg)

    def error(self, message):
        self.print_help(sys.stderr)
        self.exit(2, f'\n### Error {self.prog} -> {message}\n\n')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud
      @file: crud_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from abc import ABC, abstractmethod
from typing import List, Optional

from hspylib.core.crud.crud_entity import CrudEntity
from requests.structures import CaseInsensitiveDict


class CrudRepository(ABC):
    """TODO"""

    @abstractmethod
    def insert(self, entity: CrudEntity) -> None:
        """TODO
        param: entity
        """

    @abstractmethod
    def update(self, entity: CrudEntity) -> None:
        """TODO
        param: entity
        """

    @abstractmethod
    def delete(self, entity: CrudEntity) -> None:
        """TODO
        param: entity
        """

    @abstractmethod
    def find_all(self, filters: CaseInsensitiveDict = None) -> List[CrudEntity]:
        """TODO
        param: entity
        """

    @abstractmethod
    def find_by_id(self, entity_id: str) -> Optional[CrudEntity]:
        """TODO
        param: entity
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.security
      @file: security.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import base64
import os

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from hspylib.core.tools.preconditions import check_argument, check_state

# Please do not modify this

DEFAULT_HS_SALT = 'HsPyLib'


def encode(in_file, out_file, encoding: str = 'utf-8') -> int:
    """Encode file into base64
    :param in_file: The file to be encoded
    :param out_file: The resulting encoded file
    :param encoding: The text encoding
    """
    with open(in_file, 'r', encoding=encoding) as f_in_file:
        with open(out_file, 'w', encoding=encoding) as f_out_file:
            data = base64.b64encode(str.encode(f_in_file.read()))
            return f_out_file.write(str(data, encoding=encoding))


def decode(in_file, out_file, encoding: str = 'utf-8') -> int:
    """Decode file from base64
    :param in_file: The file to be decoded
    :param out_file: The resulting decoded file
    :param encoding: The text encoding
    """
    with open(in_file, 'r', encoding=encoding) as f_in_file:
        with open(out_file, 'w', encoding=encoding) as f_out_file:
            data = base64.b64decode(f_in_file.read())
            return f_out_file.write(str(data, encoding=encoding))


def encrypt(
    in_file: str,
    out_file: str,
    pass_phrase: str,
    salt: str = DEFAULT_HS_SALT,
    digest_algo=hashes.SHA256(),
    length=32,
    iterations=100000,
    encoding: str = 'utf-8') -> None:
    """Encrypt file using Fernet cryptography
    :param in_file: The file to be encrypted
    :param out_file: The resulting encrypted file
    :param pass_phrase: The passphrase to encrypt the file
    :param salt: TODO
    :param digest_algo: The digest encrypting algorithm
    :param length: TODO
    :param iterations: TODO
    :param encoding:
    """
    kdf = PBKDF2HMAC(
        algorithm=digest_algo,
        length=length,
        salt=salt.encode(encoding),
        iterations=iterations,
        backend=default_backend()
    )
    key = base64.urlsafe_b64encode(kdf.derive(pass_phrase.encode(encoding)))
    f = Fernet(key)
    check_argument(os.path.exists(in_file), "Input file \"{}\" does not exist", in_file)
    with open(in_file, encoding=encoding) as f_in_file:
        with open(out_file, 'w', encoding=encoding) as f_out_file:
            f_out_file.write(f.encrypt(f_in_file.read().encode(encoding)).decode(encoding))
    check_state(os.path.exists(out_file), "Unable to encrypt file \"{}\"", in_file)


def decrypt(
    in_file: str,
    out_file: str,
    pass_phrase: str,
    salt: str = DEFAULT_HS_SALT,
    digest_algo=hashes.SHA256(),
    length=32,
    iterations=100000,
    encoding: str = 'utf-8') -> None:
    """Decrypt file using Fernet cryptography
    :param in_file: The file to be decrypted
    :param out_file: The resulting decrypted file
    :param pass_phrase: The passphrase to decrypt the file
    :param salt: TODO
    :param digest_algo: The digest decrypting algorithm
    :param length: TODO
    :param iterations: TODO
    :param encoding:
    """
    kdf = PBKDF2HMAC(
        algorithm=digest_algo,
        length=length,
        salt=salt.encode(encoding),
        iterations=iterations,
        backend=default_backend()
    )
    key = base64.urlsafe_b64encode(kdf.derive(pass_phrase.encode(encoding)))
    f = Fernet(key)
    check_argument(os.path.exists(in_file), "Input file \"{}\" does not exist", in_file)
    with open(in_file, encoding=encoding) as f_in_file:
        with open(out_file, 'w', encoding=encoding) as f_out_file:
            f_out_file.write(f.decrypt(f_in_file.read().encode(encoding)).decode(encoding))
    check_state(os.path.exists(out_file), "Unable to decrypt file \"{}\"", in_file)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.db
      @file: sql_factory.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import logging as log
import os
from typing import List, Optional

from hspylib.core.crud.crud_entity import CrudEntity
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.tools.preconditions import check_argument, check_state
from requests.structures import CaseInsensitiveDict


class SqlFactory(metaclass=Singleton):
    """TODO"""

    _DEFAULT_SQL_STUBS = f'{os.path.dirname(__file__)}/sql/sql_stubs.sql'

    @staticmethod
    def _read_stubs(sql_filename: str) -> dict:
        """TODO"""
        sql_stubs = {}
        check_argument(os.path.exists(sql_filename), "Sql file was not found: {}", sql_filename)
        with open(sql_filename, encoding='utf-8') as f_stubs:
            lines = f_stubs.readlines()
            check_state(len(lines) > 0, "SQL Stub file is empty")
            lines = list(map(str.strip, lines))
            stubs = ' '.join(lines).split(';')
            check_state(
                len(stubs) >= 4, "Stub file does not have the minimum stubs for [insert, select, update, delete]")
            for stub in stubs:
                if stub:
                    key = stub.strip().partition(' ')[0].lower()
                    sql_stubs[key] = stub.strip()
        return sql_stubs

    @staticmethod
    def _join_filters(filters: CaseInsensitiveDict, join_operator: str = 'AND') -> str:
        """TODO"""
        filter_string = ''
        if filters:
            for key, value in filters.items():
                filter_string += f"{join_operator} {key} = '{value}'"
        return filter_string

    @staticmethod
    def _join_fieldset(entity: CrudEntity) -> str:
        """TODO"""
        fields = entity.to_column_set()
        field_set = ''
        for key, value in fields.items():
            field_set += f"{', ' if field_set else ''}{key} = '{value}'"
        return field_set

    def __init__(self):
        self.sql_stubs = SqlFactory._read_stubs(self._DEFAULT_SQL_STUBS)
        log.debug('%s created with %s Stubs', self.__class__.__name__, len(self.sql_stubs))

    def insert(self, entity: CrudEntity) -> Optional[str]:
        """TODO"""
        params = entity.to_values()
        sql = self.sql_stubs['insert'] \
            .replace(':columnSet', str(entity.to_columns()).replace("'", "")) \
            .replace(':valueSet', str(params))
        return sql

    def select(self, column_set: List[str] = None, filters: CaseInsensitiveDict = None) -> Optional[str]:
        """TODO"""
        sql = self.sql_stubs['select'] \
            .replace(':columnSet', '*' if not column_set else ', '.join(column_set)) \
            .replace(':filters', SqlFactory._join_filters(filters))
        return sql

    def update(self, entity: CrudEntity, filters: CaseInsensitiveDict) -> Optional[str]:
        """TODO"""
        sql = self.sql_stubs['update'] \
            .replace(':fieldSet', SqlFactory._join_fieldset(entity)) \
            .replace(':filters', SqlFactory._join_filters(filters))
        return sql

    def delete(self, filters: CaseInsensitiveDict) -> Optional[str]:
        """TODO"""
        sql = self.sql_stubs['delete'] \
            .replace(':filters', SqlFactory._join_filters(filters))
        return sql



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   main.addons.widman
      @file: widget.py
   @created: Thu, 20 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from abc import ABC, abstractmethod
from typing import List, Tuple

from hspylib.core.enums.exit_code import ExitCode
from hspylib.modules.cli.icons.font_awesome.awesome import Awesome


class Widget(ABC):
    """HSPyLib Widgets base class. All widgets must inherit from this class"""

    _USAGE_FMT = """
HSPyLib Widget: {} v{}

{}

{}
"""

    def __init__(
        self,
        icon: Awesome,
        name: str,
        tooltip: str,
        usage: str,
        version: Tuple[int, int, int]):
        self._icon = icon
        self._name = name
        self._tooltip = tooltip
        self._usage = usage
        self._version = version

    @abstractmethod
    def execute(self, args: List[str] = None) -> ExitCode:
        """Execute the widget main flow"""

    def cleanup(self) -> None:
        """Execute the widget cleanup"""

    def icon(self) -> Awesome:
        return self._icon

    def name(self) -> str:
        """Return the name about the widget"""
        return self._name

    def tooltip(self) -> str:
        """Return information about the widget"""
        return self._tooltip

    def version(self) -> str:
        """Return the version of the widget"""
        return str(self._version)

    def usage(self) -> str:
        """Return a usage message about the widget"""
        return self._USAGE_FMT.format(self._name, self._version, self._tooltip, self._usage)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.tools
      @file: text_tools.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import random
import re
import struct
from abc import ABC
from typing import Union

from hspylib.core.tools.commons import get_or_default, sysout


def fit_text(text: str, width: int) -> str:
    """TODO"""
    return text if len(text) <= width else text[0:width - 3] + '...'


def rand_string(choices: str, length: int) -> str:
    """TODO"""
    return ''.join(random.choices(choices, k=length))


def justified_left(string: str, width: int, fill: str = ' ') -> str:
    """TODO"""
    return string.ljust(width, fill)


def justified_center(string: str, width: int, fill: str = ' ') -> str:
    """TODO"""
    return string.center(width, fill)


def justified_right(string: str, width: int, fill: str = ' ') -> str:
    """TODO"""
    return string.rjust(width, fill)


def uppercase(string: str) -> str:
    """TODO"""
    return string.upper()


def lowercase(string: str) -> str:
    """TODO"""
    return string.lower()


def camelcase(string: str, separator: str = '_|-') -> str:
    """TODO"""
    parts = re.split(rf'{separator}+', string)
    return '_'.join([p.capitalize() for p in parts])


def snakecase(string: str) -> str:
    """TODO"""
    return string.strip().lower().replace(' ', '_').replace('-', '_')


def cut(string: str, index: int, separator: str = ' ') -> str:
    """TODO"""
    result = tuple(re.split(rf'{separator}+', string))
    return get_or_default(result, index)


def strip_escapes(string: str) -> str:
    """TODO"""
    return re.compile(r'\x1b[^m]*m').sub('', string)


def print_unicode(uni_code: Union[str, int]) -> None:
    """TODO"""
    if isinstance(uni_code, str) and re.match(r"^[a-fA-F0-9]{1,4}$", uni_code):
        hex_val = bytes.decode(struct.pack("!I", int(uni_code.zfill(4), 16)), 'utf_32_be')
        sysout(hex_val, end='')
    elif isinstance(uni_code, int):
        hex_val = bytes.decode(struct.pack("!I", uni_code), 'utf_32_be')
        sysout(hex_val, end='')
    else:
        raise TypeError(f'Invalid unicode: {uni_code}')


def end_ln(current: int, split_at: int, hit_str: str = '\n', miss_str: str = ' ') -> str:
    """TODO"""
    return hit_str if current != 0 and current % split_at == 0 else miss_str


# pylint: disable=too-few-public-methods
class TextAlignment(ABC):
    """
    Table cell text justification helper.
    """
    LEFT = justified_left
    CENTER = justified_center
    RIGHT = justified_right


# pylint: disable=too-few-public-methods
class TextCase(ABC):
    """
    Table cell text justification helper.
    """
    UPPER_CASE = uppercase
    LOWER_CASE = lowercase
    CAMEL_CASE = camelcase


if __name__ == '__main__':
    print_unicode('118')
    print_unicode('f118')
    print_unicode(0xf118)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.db.firebase
      @file: firebase_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import json
import logging as log
import uuid
from abc import abstractmethod
from typing import Optional

from hspylib.core.crud.crud_entity import CrudEntity
from hspylib.core.crud.crud_repository import CrudRepository
from hspylib.core.crud.db.firebase.firebase_config import FirebaseConfig
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.tools.preconditions import check_not_none, check_state
from hspylib.modules.fetch.fetch import delete, get, put
from requests.exceptions import HTTPError
from requests.structures import CaseInsensitiveDict


class FirebaseRepository(CrudRepository):
    """TODO"""

    def __init__(self):
        self.payload = None
        self.config = FirebaseConfig()

    def __str__(self):
        return str(self.payload)

    def to_list(self, json_string: str, filters: CaseInsensitiveDict = None) -> list:
        """TODO"""
        ret_list = []
        for value in json.loads(json_string).values():
            if not filters or all(k in value and value[k] == v for k, v in filters.items()):
                ret_list.append(self.row_to_entity(value))
        return ret_list

    def insert(self, entity: CrudEntity) -> None:
        """TODO"""
        entity.uuid = entity.uuid if entity.uuid is not None else str(uuid.uuid4())
        url = f'{self.config.base_url()}/{entity.uuid}.json'
        payload = entity.to_json()
        log.debug("Inserting firebase entry: %s into: %s", entity, url)
        response = put(url, payload)
        check_not_none(response, "Response is none")
        if response.status_code != HttpCode.OK:
            raise HTTPError(f'{response.status_code} - Unable to put into={url} with json_string={payload}')

    def update(self, entity: CrudEntity) -> None:
        """TODO"""
        url = f'{self.config.base_url()}/{entity.uuid}.json'
        payload = entity.to_json()
        log.debug('Updating firebase entry: %s into: %s', entity, url)
        response = put(url, payload)
        check_not_none(response, "Response is none")
        if response.status_code != HttpCode.OK:
            raise HTTPError(f'{response.status_code} - Unable to put into={url} with json_string={payload}')

    def delete(self, entity: CrudEntity) -> None:
        """TODO"""
        url = f'{self.config.base_url()}/{entity.uuid}.json'
        log.debug('Deleting firebase entry: %s into: %s', entity, url)
        response = delete(url)
        check_not_none(response, "Response is none")
        if response.status_code != HttpCode.OK:
            raise HTTPError(f'{response.status_code} - Unable to delete from={url}')

    def find_all(self, filters: CaseInsensitiveDict = None) -> Optional[list]:
        """TODO"""
        url = f'{self.config.base_url()}.json?orderBy="$key"'
        log.debug('Fetching firebase entries from %s', url)
        response = get(url)
        check_not_none(response, "Response is none")
        if response.status_code != HttpCode.OK:
            raise HTTPError(f'{response.status_code} - Unable to get from={url}')

        return self.to_list(response.body, filters) if response.body else []

    def find_by_id(self, entity_id: str) -> Optional[CrudEntity]:
        """TODO"""
        url = f'{self.config.base_url()}.json?orderBy="$key"&equalTo="{entity_id}"'
        log.debug('Fetching firebase entry entity_id=%s from %s', entity_id, url)
        response = get(url)
        check_not_none(response, "Response is none")
        if response.status_code != HttpCode.OK:
            raise HTTPError(f'{response.status_code} - Unable to get from={url}')
        result = self.to_list(response.body) if response.body else []
        check_state(len(result) <= 1, "Multiple results found with entity_id={}", entity_id)

        return result[0] if len(result) > 0 else None

    @abstractmethod
    def row_to_entity(self, row: dict) -> CrudEntity:
        pass

    @abstractmethod
    def database_name(self) -> str:
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.model
      @file: crud_entity.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import json
from abc import ABC
from typing import Tuple
from uuid import UUID

class CrudEntity(ABC):
    """TODO"""

    def __init__(self, entity_id: UUID = None):
        self.uuid = entity_id

    def __str__(self):
        return f"{self.__class__.__name__}:{str(self.to_dict())}"

    def __eq__(self, other):
        return isinstance(other, CrudEntity) \
               and all(item in self.to_dict().items() for item in other.to_dict().items())

    def to_dict(self) -> dict:
        """TODO"""

        ret_dict = {}
        for key, value in self.__dict__.items():
            if isinstance(value, bool):
                ret_dict[key] = bool(value)
            elif isinstance(value, int):
                ret_dict[key] = int(value)
            elif isinstance(value, float):
                ret_dict[key] = float(value)
            elif isinstance(value, (str, UUID)):
                ret_dict[key] = str(value)
            else:
                ret_dict[key] = value.__dict__() if value else {}
        return ret_dict

    def to_json(self) -> str:
        """TODO"""
        dict_obj = self.to_dict()
        json_str = json.dumps(dict_obj)
        return json_str

    def to_columns(self) -> Tuple[str]:
        """TODO"""
        cols = []
        for key in self.__dict__:
            if not key.startswith('_'):
                cols.append(key.replace("'", "").upper())
        return tuple(cols)

    def to_column_set(self) -> dict:
        """TODO"""
        fields = {}
        for key, value in self.__dict__.items():
            if not key.startswith('_'):
                fields[key.replace("'", "").upper()] = f"{str(value)}"
        return fields

    def to_values(self) -> Tuple[str]:
        """TODO"""
        values = []
        for key, value in self.__dict__.items():
            if not key.startswith('_'):
                values.append(str(value))
        return tuple(values)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.config
      @file: app_config.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import logging as log
import os
from textwrap import dedent
from typing import Any, Optional

from hspylib.core.config.properties import Properties
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.tools.commons import run_dir
from hspylib.core.tools.preconditions import check_argument


class AppConfigs(metaclass=Singleton):
    """Holds all of the application configurations (properties)"""

    DISPLAY_FORMAT = dedent("""
    AppConfigs
      |-Working dir = {}
      |-Resources dir = {}
      |-Properties:
       \\-{}
    """)

    def __init__(self, resource_dir: str, log_dir: str = None):

        check_argument(os.path.exists(resource_dir), "Unable to locate resources dir: {}", resource_dir)
        self._resource_dir = resource_dir

        self._app_properties = Properties(load_dir=resource_dir)

        log.info(self)

    def __str__(self):
        return '\n{}{}{}'.format(
            '-=' * 40,
            self.DISPLAY_FORMAT.format(
                str(run_dir()),
                str(self._resource_dir),
                str(self._app_properties).replace('\n', '\n   |-')
                if len(self._app_properties) > 0 else ''
            ),
            '-=' * 40
        )

    def __repr__(self):
        return str(self)

    def __getitem__(self, item: str) -> Any:
        return self.get(item)

    def __len__(self) -> int:
        return len(self._app_properties)

    def resource_dir(self) -> Optional[str]:
        """TODO"""
        return self._resource_dir

    def log_dir(self) -> Optional[str]:
        """TODO"""
        return self._log_dir

    def log_file(self) -> Optional[str]:
        """TODO"""
        return self._log_file

    def get(self, property_name: str) -> Optional[str]:
        """TODO"""
        env_value = os.environ.get(Properties.environ_name(property_name))
        return str(env_value) if env_value else self._app_properties.get(property_name)

    def get_int(self, property_name: str) -> Optional[int]:
        """TODO"""
        env = os.environ.get(Properties.environ_name(property_name))
        return int(env) if env else self._app_properties.get_int(property_name)

    def get_float(self, property_name: str) -> Optional[float]:
        """TODO"""
        env = os.environ.get(Properties.environ_name(property_name))
        return float(env) if env else self._app_properties.get_float(property_name)

    def get_bool(self, property_name: str) -> Optional[bool]:
        """TODO"""
        env = os.environ.get(Properties.environ_name(property_name))
        return bool(env) if env else self._app_properties.get_bool(property_name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.fetch
      @file: fetch.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from typing import Any, List, Optional, Tuple, Union

import requests
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.enums.http_method import HttpMethod
from hspylib.core.tools.commons import sysout
from hspylib.modules.fetch.http_response import HttpResponse
from requests import exceptions as ex


def fetch(
    url: str,
    method: HttpMethod = HttpMethod.GET,
    headers: List[str] = None,
    body: Optional[Any] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10) -> Optional[HttpResponse]:
    """ Do a request specified by method and according to parameters.
    :param url: The url to make the request.
    :param method: The http method to be used [ GET, HEAD, POST, PUT, PATCH, DELETE ].
    :param headers: The http request headers.
    :param body: The http request body (payload).
    :param silent: Omits all informational messages.
    :param timeout: How many seconds to wait for the server to send data or connect before giving up.
    :return:
    """

    url = url if url and url.startswith("http") else f'http://{url}'
    if not silent:
        sysout(f"Fetching: "
               f"method={method} table_headers={headers if headers else '[]'} "
               f"body={body if body else '{}'} url={url} ...")

    response = requests.request(
        url=url,
        method=method.name,
        headers=headers,
        data=body,
        timeout=timeout,
        verify=False)

    return HttpResponse.of(response)


def head(
    url: str,
    headers: List[str] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10) -> Optional[HttpResponse]:
    """Do HEAD request and according to parameters."""

    return fetch(
        url=url,
        method=HttpMethod.HEAD,
        headers=headers,
        silent=silent,
        timeout=timeout)


def get(
    url: str,
    headers: List[str] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10) -> Optional[HttpResponse]:
    """Do GET request and according to parameters."""

    return fetch(
        url=url,
        headers=headers,
        silent=silent,
        timeout=timeout)


def delete(
    url: str,
    headers: List[str] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10) -> Optional[HttpResponse]:
    """Do DELETE request and according to parameters."""

    return fetch(
        url=url,
        method=HttpMethod.DELETE,
        headers=headers,
        silent=silent,
        timeout=timeout)


def post(
    url: str,
    body=None,
    headers: List[str] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10) -> Optional[HttpResponse]:
    """Do POST request and according to parameters."""

    return fetch(
        url=url,
        method=HttpMethod.POST,
        headers=headers,
        body=body,
        silent=silent,
        timeout=timeout)


def put(
    url: str,
    body=None,
    headers: List[str] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10) -> Optional[HttpResponse]:
    """Do PUT request and according to parameters."""

    return fetch(
        url=url,
        method=HttpMethod.PUT,
        headers=headers,
        body=body,
        silent=silent,
        timeout=timeout)


def patch(
    url: str,
    body=None,
    headers: List[str] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10) -> Optional[HttpResponse]:
    """Do PATCH request and according to parameters."""

    return fetch(
        url=url,
        method=HttpMethod.PATCH,
        headers=headers,
        body=body,
        silent=silent,
        timeout=timeout)


def is_reachable(
    url: str,
    timeout: Union[float, Tuple[float, float]] = 1) -> bool:
    """Check if the specified url is reachable"""

    try:
        resp = fetch(url=url, method=HttpMethod.HEAD, timeout=timeout)
        return HttpCode.OK.value <= resp.status_code.value < HttpCode.MULTIPLE_CHOICES.value
    except (ex.ConnectTimeout, ex.ConnectionError, ex.ReadTimeout, ex.InvalidURL):
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.db.mysql
      @file: mysql_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import logging as log
import uuid
from abc import abstractmethod
from typing import List, Optional, Tuple

import pymysql
from hspylib.core.crud.crud_entity import CrudEntity
from hspylib.core.crud.db.db_repository import DBRepository
from hspylib.core.crud.db.sql_factory import SqlFactory
from hspylib.core.exception.exceptions import NotConnectedError
from pymysql.err import OperationalError, ProgrammingError
from requests.structures import CaseInsensitiveDict as SqlFilter


class MySqlRepository(DBRepository):
    """TODO"""

    _connection_cache = {}

    def __init__(self):
        super().__init__()
        self._connector = None
        self._cursor = None
        self._sql_factory = SqlFactory()

    def is_connected(self) -> bool:
        """TODO"""
        return self._connector is not None

    def connect(self) -> None:
        """TODO"""
        if not self.is_connected():
            cache_key = self.__str__()
            if cache_key in self._connection_cache:
                self._connector = self._connection_cache[cache_key]
                self._cursor = self._connector.cursor()
                if not self.is_connected():
                    raise ConnectionError("Not connected to the database")
            else:
                db_url = f"{self.user}@{self.hostname}:{self.port}/{self.database}"
                try:
                    self._connector = pymysql.connect(
                        host=self.hostname,
                        user=self.user,
                        port=self.port,
                        password=self.password,
                        database=self.database)
                    if not self.is_connected():
                        raise ConnectionError(f"Unable to connect to {db_url}")
                    self._cursor = self._connector.cursor()
                    log.debug("Connection to %s established", db_url)
                    self._connection_cache[cache_key] = self._connector
                except (OperationalError, ConnectionRefusedError) as err:
                    raise ConnectionError(f"Unable to connect to {db_url}") from err

    def disconnect(self) -> None:
        """TODO"""
        if self.is_connected():
            cache_key = self.__str__()
            self._connector.close()
            del self._connector
            del self._connection_cache[cache_key]
            log.debug('Disconnected from %s', str(self))
        else:
            raise NotConnectedError('Not connected to database.')

    def insert(self, entity: CrudEntity) -> None:
        """TODO"""
        if self.is_connected():
            entity.uuid = entity.uuid if entity.uuid is not None else str(uuid.uuid4())
            stm = self._sql_factory \
                .insert(entity) \
                .replace(':tableName', self.table_name())
            log.debug("Executing SQL statement: %s", stm)
            self.execute(stm, True)
        else:
            raise NotConnectedError('Not connected to database.')

    def update(self, entity: CrudEntity) -> None:
        """TODO"""
        if self.is_connected():
            stm = self._sql_factory \
                .update(entity, filters=SqlFilter({"UUID": f'{entity.uuid}'})) \
                .replace(':tableName', self.table_name())
            log.debug('Executing SQL statement: %s', stm)
            self.execute(stm, True)
        else:
            raise NotConnectedError('Not connected to database.')

    def delete(self, entity: CrudEntity) -> None:
        """TODO"""
        if self.is_connected():
            stm = self._sql_factory \
                .delete(filters=SqlFilter({"UUID": f'{entity.uuid}'})) \
                .replace(':tableName', self.table_name())
            log.debug('Executing SQL statement: %s', stm)
            self.execute(stm, True)
        else:
            raise NotConnectedError('Not connected to database.')

    def find_all(  # pylint: disable=arguments-differ
        self,
        column_set: List[str] = None,
        sql_filters: SqlFilter = None) -> Optional[list]:
        """TODO"""

        if self.is_connected():
            stm = self._sql_factory \
                .select(column_set=column_set, filters=sql_filters) \
                .replace(':tableName', self.table_name())
            log.debug('Executing SQL statement: %s', stm)
            self.execute(stm, True)
            result = self._cursor.fetchall()
            return list(map(self.row_to_entity, result)) if result else None

        raise NotConnectedError('Not connected to database.')

    def find_by_id(  # pylint: disable=arguments-differ
        self,
        column_set: List[str] = None,
        entity_id: str = None) -> Optional[CrudEntity]:
        """TODO"""

        if self.is_connected():
            if entity_id:
                stm = self._sql_factory \
                    .select(column_set=column_set, filters=SqlFilter({"UUID": f'{entity_id}'})) \
                    .replace(':tableName', self.table_name())
                log.debug('Executing SQL statement: %s', stm)
                self.execute(stm, True)
                result = self._cursor.fetchall()
                if len(result) > 1:
                    raise ProgrammingError(f'Multiple results found {len(result)}')
                return self.row_to_entity(result[0]) if len(result) > 0 else None
            return None

        raise NotConnectedError('Not connected to database.')

    def execute(self, sql_statement: str, auto_commit: bool, *params) -> None:
        """TODO"""
        if self.is_connected():
            log.debug("Executing SQL statement: %s with params: [%s]", sql_statement, ', '.join(params))
            self._cursor.execute(sql_statement, *params)
            if auto_commit:
                self.commit()
        else:
            raise NotConnectedError('Not connected to database.')

    def commit(self) -> None:
        """TODO"""
        log.debug('Committing database changes')
        self._connector.commit()

    def rollback(self) -> None:
        """TODO"""
        log.debug('Rolling back database changes')
        self._connector.rollback()

    @abstractmethod
    def row_to_entity(self, row: Tuple) -> CrudEntity:
        pass

    @abstractmethod
    def table_name(self) -> str:
        pass



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.39
#
# Package: main.hspylib.modules.cli.icons.font_awesome
"""Package initialization."""

__all__ = [
  'awesome', 
  'dashboard_icons', 
  'form_icons', 
  'widget_icons'
]
__version__ = '0.11.39'



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.39
#
# Package: main.hspylib.modules.qt.promotions
"""Package initialization."""

__all__ = [
  'hcombobox', 
  'hconsole', 
  'hframe', 
  'hlabel', 
  'hlistwidget', 
  'htablemodel', 
  'htableview'
]
__version__ = '0.11.39'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.tools
      @file: preconditions.py
   @created: Fri, 16 Jul 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from typing import Tuple, TypeVar

from hspylib.core.exception.exceptions import InvalidArgumentError, InvalidStateError

T = TypeVar('T')


def check_argument(expression: bool, error_message: str = None, *args) -> bool:
    """Ensures the truth of an expression involving one or more parameters to the calling method."""
    if not expression:
        raise InvalidArgumentError(
            error_message.format(*args) if error_message else 'Precondition failed: Invalid argument')

    return expression


def check_state(expression: bool, error_message: str = None, *args) -> bool:
    """Ensures the truth of an expression involving the state of the calling instance, but not involving
    any parameters to the calling method."""
    if not expression:
        raise InvalidStateError(
            error_message.format(*args) if error_message else 'Precondition failed: Invalid state')

    return expression


def check_not_none(reference: T, error_message: str = None, *args) -> T:
    """Ensures that an object reference passed as a parameter to the calling method is not None."""
    if reference is None:
        raise TypeError(
            error_message.format(*args) if error_message else 'Precondition failed: Null reference')
    return reference


def check_element_index(index: int, size: int, desc: str = None) -> int:
    """Ensures that index specifies a valid element in an array, list or string of size size."""

    if size < 0:
        raise InvalidArgumentError('Size is negative')
    if index < 0 or index >= size:
        raise IndexError(desc or 'Precondition failed: Index is negative or out of bounds')

    return index


def check_index_in_range(start: int, end: int, size: int, desc: str = None) -> Tuple[int, int]:
    """Ensures that start and end specify a valid positions in an array, list or string of size size, and
    are in order."""
    if size < 0:
        raise InvalidArgumentError('Size is negative')
    if start < 0 or end < 0 or start >= size or end >= size:
        raise IndexError(desc or 'Precondition failed: Index is negative or greater than size')
    if end < start:
        raise IndexError(desc or 'Precondition failed: End is less than start')

    return start, end



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.39
#
# Package: main.hspylib.modules.cli
"""Package initialization."""

__all__ = [
  'application', 
  'icons', 
  'keyboard', 
  'tui', 
  'vt100'
]
__version__ = '0.11.39'



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.39
#
# Package: main.hspylib.addons.appman
"""Package initialization."""

__all__ = [
  'app_extension', 
  'app_type', 
  'appman', 
  'templates'
]
__version__ = '0.11.39'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.exception
      @file: exceptions.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import logging as log

class HSBaseException(Exception):
    """This is a generic exception and should not be raised. It may be inherited instead"""

    def __init__(self, message: str, cause: Exception = None):
        fmt_msg = f'{message}' + (' => {str(cause)}' if cause else '')
        super().__init__(fmt_msg)
        log.error(fmt_msg)


class EntityNotFoundError(HSBaseException):
    """Raised when an entity is not found"""


class InputAbortedError(HSBaseException):
    """Raised when an input method is aborted"""


class ProgrammingError(HSBaseException):
    """Exception raised for programming errors, e.g. table not found
    or already exists, syntax error in the SQL statement, wrong number
    of parameters specified, etc."""


class InvalidOptionError(HSBaseException):
    """Raised when an invalid option is received by the application"""


class NotConnectedError(HSBaseException):
    """Raised when an attempt to execute without being connected but a connection is required"""


class WidgetNotFoundError(HSBaseException):
    """Raised when widget is not found on any of the widget paths"""


class WidgetExecutionError(HSBaseException):
    """Raised when widget failed to execute"""


class KeyboardInputError(HSBaseException):
    """Raised when an invalid keystroke input is provided"""


class UnsupportedSchemaError(HSBaseException):
    """Raised when an invalid schema is provided """


class SchemaRegistryError(HSBaseException):
    """Raised when schema registration or API access fails"""


class InvalidInputError(Exception):
    """Raised when an invalid input is provided """


class InvalidArgumentError(Exception):
    """Raised when an invalid argument is received by the application"""


class InvalidStateError(Exception):
    """Raised when an invalid state is detected by the application"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.icons.font_awesome
      @file: awesome.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration
from hspylib.core.tools.commons import sysout
from hspylib.core.tools.text_tools import end_ln, print_unicode

class Awesome(Enumeration):
    """
        Font awesome codes
        Full list of font awesome icons can be found here:
          - https://fontawesome.com/cheatsheet?from=io
    """

    @staticmethod
    def demo_unicodes() -> None:
        n = 0
        st_base = [f'F{x:03X}' for x in range(0, 4095)]
        for h in st_base:
            print_unicode(h)
            sysout(f' {h:4}', end=end_ln(n, 10))
            n += 1

    @classmethod
    def demo_icons(cls) -> None:
        i = 0
        for n in cls.values():
            sysout(f'{n:2}', end=end_ln(n, 10))
            i += 1

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self):
        return str(self)

    def placeholder(self) -> str:
        return f":{self.name}:"


if __name__ == '__main__':
    Awesome.demo_unicodes()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.enum
      @file: http_method.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration

class HttpMethod(Enumeration):
    """TODO"""

    # @formatter:off
    OPTIONS     = 'options'
    HEAD        = 'head'
    GET         = 'get'
    POST        = 'post'
    PUT         = 'put'
    PATCH       = 'patch'
    DELETE      = 'delete'
    # @formatter:on
